#define _CFFI_

/* We try to define Py_LIMITED_API before including Python.h.

   Mess: we can only define it if Py_DEBUG, Py_TRACE_REFS and
   Py_REF_DEBUG are not defined.  This is a best-effort approximation:
   we can learn about Py_DEBUG from pyconfig.h, but it is unclear if
   the same works for the other two macros.  Py_DEBUG implies them,
   but not the other way around.

   The implementation is messy (issue #350): on Windows, with _MSC_VER,
   we have to define Py_LIMITED_API even before including pyconfig.h.
   In that case, we guess what pyconfig.h will do to the macros above,
   and check our guess after the #include.

   Note that on Windows, with CPython 3.x, you need >= 3.5 and virtualenv
   version >= 16.0.0.  With older versions of either, you don't get a
   copy of PYTHON3.DLL in the virtualenv.  We can't check the version of
   CPython *before* we even include pyconfig.h.  ffi.set_source() puts
   a ``#define _CFFI_NO_LIMITED_API'' at the start of this file if it is
   running on Windows < 3.5, as an attempt at fixing it, but that's
   arguably wrong because it may not be the target version of Python.
   Still better than nothing I guess.  As another workaround, you can
   remove the definition of Py_LIMITED_API here.

   See also 'py_limited_api' in cffi/setuptools_ext.py.
*/
#if !defined(_CFFI_USE_EMBEDDING) && !defined(Py_LIMITED_API)
#  ifdef _MSC_VER
#    if !defined(_DEBUG) && !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#    include <pyconfig.h>
     /* sanity-check: Py_LIMITED_API will cause crashes if any of these
        are also defined.  Normally, the Python file PC/pyconfig.h does not
        cause any of these to be defined, with the exception that _DEBUG
        causes Py_DEBUG.  Double-check that. */
#    ifdef Py_LIMITED_API
#      if defined(Py_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_DEBUG, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_TRACE_REFS)
#        error "pyconfig.h unexpectedly defines Py_TRACE_REFS, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_REF_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_REF_DEBUG, but Py_LIMITED_API is set"
#      endif
#    endif
#  else
#    include <pyconfig.h>
#    if !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#  endif
#endif

#include <Python.h>
#ifdef __cplusplus
extern "C" {
#endif
#include <stddef.h>

/* This part is from file 'cffi/parse_c_type.h'.  It is copied at the
   beginning of C sources generated by CFFI's ffi.set_source(). */

typedef void *_cffi_opcode_t;

#define _CFFI_OP(opcode, arg)   (_cffi_opcode_t)(opcode | (((uintptr_t)(arg)) << 8))
#define _CFFI_GETOP(cffi_opcode)    ((unsigned char)(uintptr_t)cffi_opcode)
#define _CFFI_GETARG(cffi_opcode)   (((intptr_t)cffi_opcode) >> 8)

#define _CFFI_OP_PRIMITIVE       1
#define _CFFI_OP_POINTER         3
#define _CFFI_OP_ARRAY           5
#define _CFFI_OP_OPEN_ARRAY      7
#define _CFFI_OP_STRUCT_UNION    9
#define _CFFI_OP_ENUM           11
#define _CFFI_OP_FUNCTION       13
#define _CFFI_OP_FUNCTION_END   15
#define _CFFI_OP_NOOP           17
#define _CFFI_OP_BITFIELD       19
#define _CFFI_OP_TYPENAME       21
#define _CFFI_OP_CPYTHON_BLTN_V 23   // varargs
#define _CFFI_OP_CPYTHON_BLTN_N 25   // noargs
#define _CFFI_OP_CPYTHON_BLTN_O 27   // O  (i.e. a single arg)
#define _CFFI_OP_CONSTANT       29
#define _CFFI_OP_CONSTANT_INT   31
#define _CFFI_OP_GLOBAL_VAR     33
#define _CFFI_OP_DLOPEN_FUNC    35
#define _CFFI_OP_DLOPEN_CONST   37
#define _CFFI_OP_GLOBAL_VAR_F   39
#define _CFFI_OP_EXTERN_PYTHON  41

#define _CFFI_PRIM_VOID          0
#define _CFFI_PRIM_BOOL          1
#define _CFFI_PRIM_CHAR          2
#define _CFFI_PRIM_SCHAR         3
#define _CFFI_PRIM_UCHAR         4
#define _CFFI_PRIM_SHORT         5
#define _CFFI_PRIM_USHORT        6
#define _CFFI_PRIM_INT           7
#define _CFFI_PRIM_UINT          8
#define _CFFI_PRIM_LONG          9
#define _CFFI_PRIM_ULONG        10
#define _CFFI_PRIM_LONGLONG     11
#define _CFFI_PRIM_ULONGLONG    12
#define _CFFI_PRIM_FLOAT        13
#define _CFFI_PRIM_DOUBLE       14
#define _CFFI_PRIM_LONGDOUBLE   15

#define _CFFI_PRIM_WCHAR        16
#define _CFFI_PRIM_INT8         17
#define _CFFI_PRIM_UINT8        18
#define _CFFI_PRIM_INT16        19
#define _CFFI_PRIM_UINT16       20
#define _CFFI_PRIM_INT32        21
#define _CFFI_PRIM_UINT32       22
#define _CFFI_PRIM_INT64        23
#define _CFFI_PRIM_UINT64       24
#define _CFFI_PRIM_INTPTR       25
#define _CFFI_PRIM_UINTPTR      26
#define _CFFI_PRIM_PTRDIFF      27
#define _CFFI_PRIM_SIZE         28
#define _CFFI_PRIM_SSIZE        29
#define _CFFI_PRIM_INT_LEAST8   30
#define _CFFI_PRIM_UINT_LEAST8  31
#define _CFFI_PRIM_INT_LEAST16  32
#define _CFFI_PRIM_UINT_LEAST16 33
#define _CFFI_PRIM_INT_LEAST32  34
#define _CFFI_PRIM_UINT_LEAST32 35
#define _CFFI_PRIM_INT_LEAST64  36
#define _CFFI_PRIM_UINT_LEAST64 37
#define _CFFI_PRIM_INT_FAST8    38
#define _CFFI_PRIM_UINT_FAST8   39
#define _CFFI_PRIM_INT_FAST16   40
#define _CFFI_PRIM_UINT_FAST16  41
#define _CFFI_PRIM_INT_FAST32   42
#define _CFFI_PRIM_UINT_FAST32  43
#define _CFFI_PRIM_INT_FAST64   44
#define _CFFI_PRIM_UINT_FAST64  45
#define _CFFI_PRIM_INTMAX       46
#define _CFFI_PRIM_UINTMAX      47
#define _CFFI_PRIM_FLOATCOMPLEX 48
#define _CFFI_PRIM_DOUBLECOMPLEX 49
#define _CFFI_PRIM_CHAR16       50
#define _CFFI_PRIM_CHAR32       51

#define _CFFI__NUM_PRIM         52
#define _CFFI__UNKNOWN_PRIM           (-1)
#define _CFFI__UNKNOWN_FLOAT_PRIM     (-2)
#define _CFFI__UNKNOWN_LONG_DOUBLE    (-3)

#define _CFFI__IO_FILE_STRUCT         (-1)


struct _cffi_global_s {
    const char *name;
    void *address;
    _cffi_opcode_t type_op;
    void *size_or_direct_fn;  // OP_GLOBAL_VAR: size, or 0 if unknown
                              // OP_CPYTHON_BLTN_*: addr of direct function
};

struct _cffi_getconst_s {
    unsigned long long value;
    const struct _cffi_type_context_s *ctx;
    int gindex;
};

struct _cffi_struct_union_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_STRUCT_UNION
    int flags;               // _CFFI_F_* flags below
    size_t size;
    int alignment;
    int first_field_index;   // -> _cffi_fields array
    int num_fields;
};
#define _CFFI_F_UNION         0x01   // is a union, not a struct
#define _CFFI_F_CHECK_FIELDS  0x02   // complain if fields are not in the
                                     // "standard layout" or if some are missing
#define _CFFI_F_PACKED        0x04   // for CHECK_FIELDS, assume a packed struct
#define _CFFI_F_EXTERNAL      0x08   // in some other ffi.include()
#define _CFFI_F_OPAQUE        0x10   // opaque

struct _cffi_field_s {
    const char *name;
    size_t field_offset;
    size_t field_size;
    _cffi_opcode_t field_type_op;
};

struct _cffi_enum_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_ENUM
    int type_prim;           // _CFFI_PRIM_xxx
    const char *enumerators; // comma-delimited string
};

struct _cffi_typename_s {
    const char *name;
    int type_index;   /* if opaque, points to a possibly artificial
                         OP_STRUCT which is itself opaque */
};

struct _cffi_type_context_s {
    _cffi_opcode_t *types;
    const struct _cffi_global_s *globals;
    const struct _cffi_field_s *fields;
    const struct _cffi_struct_union_s *struct_unions;
    const struct _cffi_enum_s *enums;
    const struct _cffi_typename_s *typenames;
    int num_globals;
    int num_struct_unions;
    int num_enums;
    int num_typenames;
    const char *const *includes;
    int num_types;
    int flags;      /* future extension */
};

struct _cffi_parse_info_s {
    const struct _cffi_type_context_s *ctx;
    _cffi_opcode_t *output;
    unsigned int output_size;
    size_t error_location;
    const char *error_message;
};

struct _cffi_externpy_s {
    const char *name;
    size_t size_of_result;
    void *reserved1, *reserved2;
};

#ifdef _CFFI_INTERNAL
static int parse_c_type(struct _cffi_parse_info_s *info, const char *input);
static int search_in_globals(const struct _cffi_type_context_s *ctx,
                             const char *search, size_t search_len);
static int search_in_struct_unions(const struct _cffi_type_context_s *ctx,
                                   const char *search, size_t search_len);
#endif

/* this block of #ifs should be kept exactly identical between
   c/_cffi_backend.c, cffi/vengine_cpy.py, cffi/vengine_gen.py
   and cffi/_cffi_include.h */
#if defined(_MSC_VER)
# include <malloc.h>   /* for alloca() */
# if _MSC_VER < 1600   /* MSVC < 2010 */
   typedef __int8 int8_t;
   typedef __int16 int16_t;
   typedef __int32 int32_t;
   typedef __int64 int64_t;
   typedef unsigned __int8 uint8_t;
   typedef unsigned __int16 uint16_t;
   typedef unsigned __int32 uint32_t;
   typedef unsigned __int64 uint64_t;
   typedef __int8 int_least8_t;
   typedef __int16 int_least16_t;
   typedef __int32 int_least32_t;
   typedef __int64 int_least64_t;
   typedef unsigned __int8 uint_least8_t;
   typedef unsigned __int16 uint_least16_t;
   typedef unsigned __int32 uint_least32_t;
   typedef unsigned __int64 uint_least64_t;
   typedef __int8 int_fast8_t;
   typedef __int16 int_fast16_t;
   typedef __int32 int_fast32_t;
   typedef __int64 int_fast64_t;
   typedef unsigned __int8 uint_fast8_t;
   typedef unsigned __int16 uint_fast16_t;
   typedef unsigned __int32 uint_fast32_t;
   typedef unsigned __int64 uint_fast64_t;
   typedef __int64 intmax_t;
   typedef unsigned __int64 uintmax_t;
# else
#  include <stdint.h>
# endif
# if _MSC_VER < 1800   /* MSVC < 2013 */
#  ifndef __cplusplus
    typedef unsigned char _Bool;
#  endif
# endif
#else
# include <stdint.h>
# if (defined (__SVR4) && defined (__sun)) || defined(_AIX) || defined(__hpux)
#  include <alloca.h>
# endif
#endif

#ifdef __GNUC__
# define _CFFI_UNUSED_FN  __attribute__((unused))
#else
# define _CFFI_UNUSED_FN  /* nothing */
#endif

#ifdef __cplusplus
# ifndef _Bool
   typedef bool _Bool;   /* semi-hackish: C++ has no _Bool; bool is builtin */
# endif
#endif

/**********  CPython-specific section  **********/
#ifndef PYPY_VERSION


#if PY_MAJOR_VERSION >= 3
# define PyInt_FromLong PyLong_FromLong
#endif

#define _cffi_from_c_double PyFloat_FromDouble
#define _cffi_from_c_float PyFloat_FromDouble
#define _cffi_from_c_long PyInt_FromLong
#define _cffi_from_c_ulong PyLong_FromUnsignedLong
#define _cffi_from_c_longlong PyLong_FromLongLong
#define _cffi_from_c_ulonglong PyLong_FromUnsignedLongLong
#define _cffi_from_c__Bool PyBool_FromLong

#define _cffi_to_c_double PyFloat_AsDouble
#define _cffi_to_c_float PyFloat_AsDouble

#define _cffi_from_c_int(x, type)                                        \
    (((type)-1) > 0 ? /* unsigned */                                     \
        (sizeof(type) < sizeof(long) ?                                   \
            PyInt_FromLong((long)x) :                                    \
         sizeof(type) == sizeof(long) ?                                  \
            PyLong_FromUnsignedLong((unsigned long)x) :                  \
            PyLong_FromUnsignedLongLong((unsigned long long)x)) :        \
        (sizeof(type) <= sizeof(long) ?                                  \
            PyInt_FromLong((long)x) :                                    \
            PyLong_FromLongLong((long long)x)))

#define _cffi_to_c_int(o, type)                                          \
    ((type)(                                                             \
     sizeof(type) == 1 ? (((type)-1) > 0 ? (type)_cffi_to_c_u8(o)        \
                                         : (type)_cffi_to_c_i8(o)) :     \
     sizeof(type) == 2 ? (((type)-1) > 0 ? (type)_cffi_to_c_u16(o)       \
                                         : (type)_cffi_to_c_i16(o)) :    \
     sizeof(type) == 4 ? (((type)-1) > 0 ? (type)_cffi_to_c_u32(o)       \
                                         : (type)_cffi_to_c_i32(o)) :    \
     sizeof(type) == 8 ? (((type)-1) > 0 ? (type)_cffi_to_c_u64(o)       \
                                         : (type)_cffi_to_c_i64(o)) :    \
     (Py_FatalError("unsupported size for type " #type), (type)0)))

#define _cffi_to_c_i8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[1])
#define _cffi_to_c_u8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[2])
#define _cffi_to_c_i16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[3])
#define _cffi_to_c_u16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[4])
#define _cffi_to_c_i32                                                   \
                 ((int(*)(PyObject *))_cffi_exports[5])
#define _cffi_to_c_u32                                                   \
                 ((unsigned int(*)(PyObject *))_cffi_exports[6])
#define _cffi_to_c_i64                                                   \
                 ((long long(*)(PyObject *))_cffi_exports[7])
#define _cffi_to_c_u64                                                   \
                 ((unsigned long long(*)(PyObject *))_cffi_exports[8])
#define _cffi_to_c_char                                                  \
                 ((int(*)(PyObject *))_cffi_exports[9])
#define _cffi_from_c_pointer                                             \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[10])
#define _cffi_to_c_pointer                                               \
    ((char *(*)(PyObject *, struct _cffi_ctypedescr *))_cffi_exports[11])
#define _cffi_get_struct_layout                                          \
    not used any more
#define _cffi_restore_errno                                              \
    ((void(*)(void))_cffi_exports[13])
#define _cffi_save_errno                                                 \
    ((void(*)(void))_cffi_exports[14])
#define _cffi_from_c_char                                                \
    ((PyObject *(*)(char))_cffi_exports[15])
#define _cffi_from_c_deref                                               \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[16])
#define _cffi_to_c                                                       \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[17])
#define _cffi_from_c_struct                                              \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[18])
#define _cffi_to_c_wchar_t                                               \
    ((_cffi_wchar_t(*)(PyObject *))_cffi_exports[19])
#define _cffi_from_c_wchar_t                                             \
    ((PyObject *(*)(_cffi_wchar_t))_cffi_exports[20])
#define _cffi_to_c_long_double                                           \
    ((long double(*)(PyObject *))_cffi_exports[21])
#define _cffi_to_c__Bool                                                 \
    ((_Bool(*)(PyObject *))_cffi_exports[22])
#define _cffi_prepare_pointer_call_argument                              \
    ((Py_ssize_t(*)(struct _cffi_ctypedescr *,                           \
                    PyObject *, char **))_cffi_exports[23])
#define _cffi_convert_array_from_object                                  \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[24])
#define _CFFI_CPIDX  25
#define _cffi_call_python                                                \
    ((void(*)(struct _cffi_externpy_s *, char *))_cffi_exports[_CFFI_CPIDX])
#define _cffi_to_c_wchar3216_t                                           \
    ((int(*)(PyObject *))_cffi_exports[26])
#define _cffi_from_c_wchar3216_t                                         \
    ((PyObject *(*)(int))_cffi_exports[27])
#define _CFFI_NUM_EXPORTS 28

struct _cffi_ctypedescr;

static void *_cffi_exports[_CFFI_NUM_EXPORTS];

#define _cffi_type(index)   (                           \
    assert((((uintptr_t)_cffi_types[index]) & 1) == 0), \
    (struct _cffi_ctypedescr *)_cffi_types[index])

static PyObject *_cffi_init(const char *module_name, Py_ssize_t version,
                            const struct _cffi_type_context_s *ctx)
{
    PyObject *module, *o_arg, *new_module;
    void *raw[] = {
        (void *)module_name,
        (void *)version,
        (void *)_cffi_exports,
        (void *)ctx,
    };

    module = PyImport_ImportModule("_cffi_backend");
    if (module == NULL)
        goto failure;

    o_arg = PyLong_FromVoidPtr((void *)raw);
    if (o_arg == NULL)
        goto failure;

    new_module = PyObject_CallMethod(
        module, (char *)"_init_cffi_1_0_external_module", (char *)"O", o_arg);

    Py_DECREF(o_arg);
    Py_DECREF(module);
    return new_module;

  failure:
    Py_XDECREF(module);
    return NULL;
}


#ifdef HAVE_WCHAR_H
typedef wchar_t _cffi_wchar_t;
#else
typedef uint16_t _cffi_wchar_t;   /* same random pick as _cffi_backend.c */
#endif

_CFFI_UNUSED_FN static uint16_t _cffi_to_c_char16_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return (uint16_t)_cffi_to_c_wchar_t(o);
    else
        return (uint16_t)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char16_t(uint16_t x)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

_CFFI_UNUSED_FN static int _cffi_to_c_char32_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return (int)_cffi_to_c_wchar_t(o);
    else
        return (int)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char32_t(unsigned int x)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

union _cffi_union_alignment_u {
    unsigned char m_char;
    unsigned short m_short;
    unsigned int m_int;
    unsigned long m_long;
    unsigned long long m_longlong;
    float m_float;
    double m_double;
    long double m_longdouble;
};

struct _cffi_freeme_s {
    struct _cffi_freeme_s *next;
    union _cffi_union_alignment_u alignment;
};

_CFFI_UNUSED_FN static int
_cffi_convert_array_argument(struct _cffi_ctypedescr *ctptr, PyObject *arg,
                             char **output_data, Py_ssize_t datasize,
                             struct _cffi_freeme_s **freeme)
{
    char *p;
    if (datasize < 0)
        return -1;

    p = *output_data;
    if (p == NULL) {
        struct _cffi_freeme_s *fp = (struct _cffi_freeme_s *)PyObject_Malloc(
            offsetof(struct _cffi_freeme_s, alignment) + (size_t)datasize);
        if (fp == NULL)
            return -1;
        fp->next = *freeme;
        *freeme = fp;
        p = *output_data = (char *)&fp->alignment;
    }
    memset((void *)p, 0, (size_t)datasize);
    return _cffi_convert_array_from_object(p, ctptr, arg);
}

_CFFI_UNUSED_FN static void
_cffi_free_array_arguments(struct _cffi_freeme_s *freeme)
{
    do {
        void *p = (void *)freeme;
        freeme = freeme->next;
        PyObject_Free(p);
    } while (freeme != NULL);
}

/**********  end CPython-specific section  **********/
#else
_CFFI_UNUSED_FN
static void (*_cffi_call_python_org)(struct _cffi_externpy_s *, char *);
# define _cffi_call_python  _cffi_call_python_org
#endif


#define _cffi_array_len(array)   (sizeof(array) / sizeof((array)[0]))

#define _cffi_prim_int(size, sign)                                      \
    ((size) == 1 ? ((sign) ? _CFFI_PRIM_INT8  : _CFFI_PRIM_UINT8)  :    \
     (size) == 2 ? ((sign) ? _CFFI_PRIM_INT16 : _CFFI_PRIM_UINT16) :    \
     (size) == 4 ? ((sign) ? _CFFI_PRIM_INT32 : _CFFI_PRIM_UINT32) :    \
     (size) == 8 ? ((sign) ? _CFFI_PRIM_INT64 : _CFFI_PRIM_UINT64) :    \
     _CFFI__UNKNOWN_PRIM)

#define _cffi_prim_float(size)                                          \
    ((size) == sizeof(float) ? _CFFI_PRIM_FLOAT :                       \
     (size) == sizeof(double) ? _CFFI_PRIM_DOUBLE :                     \
     (size) == sizeof(long double) ? _CFFI__UNKNOWN_LONG_DOUBLE :       \
     _CFFI__UNKNOWN_FLOAT_PRIM)

#define _cffi_check_int(got, got_nonpos, expected)      \
    ((got_nonpos) == (expected <= 0) &&                 \
     (got) == (unsigned long long)expected)

#ifdef MS_WIN32
# define _cffi_stdcall  __stdcall
#else
# define _cffi_stdcall  /* nothing */
#endif

#ifdef __cplusplus
}
#endif

/************************************************************/


/* Declarations from PortAudio's pa_ringbuffer.h: */
typedef long ring_buffer_size_t;
typedef struct PaUtilRingBuffer
{
    ring_buffer_size_t  bufferSize;
    volatile ring_buffer_size_t writeIndex;
    volatile ring_buffer_size_t readIndex;
    ring_buffer_size_t bigMask;
    ring_buffer_size_t smallMask;
    ring_buffer_size_t elementSizeBytes;
    char* buffer;
} PaUtilRingBuffer;
ring_buffer_size_t PaUtil_InitializeRingBuffer(PaUtilRingBuffer* rbuf, ring_buffer_size_t elementSizeBytes, ring_buffer_size_t elementCount, void* dataPtr);
void PaUtil_FlushRingBuffer(PaUtilRingBuffer* rbuf);
ring_buffer_size_t PaUtil_GetRingBufferWriteAvailable(const PaUtilRingBuffer* rbuf);
ring_buffer_size_t PaUtil_GetRingBufferReadAvailable(const PaUtilRingBuffer* rbuf);
ring_buffer_size_t PaUtil_WriteRingBuffer(PaUtilRingBuffer* rbuf, const void* data, ring_buffer_size_t elementCount);
ring_buffer_size_t PaUtil_ReadRingBuffer(PaUtilRingBuffer* rbuf, void* data, ring_buffer_size_t elementCount);
ring_buffer_size_t PaUtil_GetRingBufferWriteRegions(PaUtilRingBuffer* rbuf, ring_buffer_size_t elementCount, void** dataPtr1, ring_buffer_size_t* sizePtr1, void** dataPtr2, ring_buffer_size_t* sizePtr2);
ring_buffer_size_t PaUtil_AdvanceRingBufferWriteIndex(PaUtilRingBuffer* rbuf, ring_buffer_size_t elementCount);
ring_buffer_size_t PaUtil_GetRingBufferReadRegions(PaUtilRingBuffer* rbuf, ring_buffer_size_t elementCount, void** dataPtr1, ring_buffer_size_t* sizePtr1, void** dataPtr2, ring_buffer_size_t* sizePtr2);
ring_buffer_size_t PaUtil_AdvanceRingBufferReadIndex(PaUtilRingBuffer* rbuf, ring_buffer_size_t elementCount);
/* End of pa_ringbuffer.h declarations. */
/* See ../rtmixer_build.py, the ring buffer declarations are included there */

#include <math.h>  // for llround()
#include <stdio.h>  // for printf()
#include <string.h>  // for memset()
#include <portaudio.h>
#include "rtmixer.h"

static const struct stats EMPTY_STATS;

#ifdef NDEBUG
#define CALLBACK_ASSERT(expr) ((void)(0))
#else
#define CALLBACK_ASSERT(expr) \
  do { if (!(expr)) { \
    printf("Failed assertion in audio callback: \"" #expr "\" (" __FILE__ \
        ":%i)\n", __LINE__); \
    return paAbort; \
  }} while (false)
#endif

void remove_action(struct action** addr, const struct state* state)
{
  struct action* action = *addr;
  *addr = action->next;  // Current action is removed from list
  action->next = NULL;
  ring_buffer_size_t written = PaUtil_WriteRingBuffer(state->result_q
    , &action, 1);
  if (written != 1)
  {
    // TODO: Stop callback! Unrecoverable error! Log error (in "state")?
    printf("result queue is full\n");
  }
}

void get_stats(frame_t frameCount, PaStreamCallbackFlags flags
  , struct stats* stats)
{
  if (stats->blocks == 0)
  {
    stats->min_blocksize = frameCount;
    stats->max_blocksize = frameCount;
  }
  else
  {
    if (frameCount < stats->min_blocksize)
    {
      stats->min_blocksize = frameCount;
    }
    if (frameCount > stats->max_blocksize)
    {
      stats->max_blocksize = frameCount;
    }
  }
  stats->blocks++;

  if (flags & paInputUnderflow)  { stats->input_underflows++; }
  if (flags & paInputOverflow)   { stats->input_overflows++; }
  if (flags & paOutputUnderflow) { stats->output_underflows++; }
  if (flags & paOutputOverflow)  { stats->output_overflows++; }
}

// CANCEL actions are inserted in the front, others in the back of the list.
// Linked lists of actions are supported.
void get_new_actions(struct state* state)
{
  struct action** last_action_addr = &(state->actions);
  for (struct action* new_action = NULL
      ; PaUtil_ReadRingBuffer(state->action_q, &new_action, 1)
      ;)
  {
    do
    {
      struct action* next = new_action->next;
      if (new_action->type == CANCEL)
      {
        new_action->next = state->actions;
        state->actions = new_action;
      }
      else
      {
        new_action->next = NULL;
        while (*last_action_addr)
        {
          last_action_addr = &((*last_action_addr)->next);
        }
        *last_action_addr = new_action;
      }
      new_action = next;
    }
    while (new_action);
  }
}

frame_t seconds2samples(PaTime time, double samplerate)
{
  return (frame_t) llround(time * samplerate);
}

PaTime get_relevant_time(const struct action* action
  , const PaStreamCallbackTimeInfo* timeInfo)
{
  enum actiontype type = action->type;
  if (type == CANCEL)
  {
    CALLBACK_ASSERT(action->action);
    type = action->action->type;
  }

  switch (type)
  {
    case PLAY_BUFFER:
    case PLAY_RINGBUFFER:
      return timeInfo->outputBufferDacTime;
    case RECORD_BUFFER:
    case RECORD_RINGBUFFER:
      return timeInfo->inputBufferAdcTime;
    default:
      return timeInfo->currentTime;
  }
}

int callback(const void* input, void* output, frame_t frameCount
  , const PaStreamCallbackTimeInfo* timeInfo, PaStreamCallbackFlags statusFlags
  , void* userData)
{
  struct state* state = userData;
  CALLBACK_ASSERT(state);

  memset(output, 0, sizeof(float) * state->output_channels * frameCount);

  get_stats(frameCount, statusFlags, &(state->stats));

  get_new_actions(state);

  // TODO: store min/max available space in result_q?
  // TODO: use worst case from before/after the "while" loop?

  struct action** actionaddr = &(state->actions);
  while (*actionaddr)
  {
    struct action* const action = *actionaddr;

    PaTime time = get_relevant_time(action, timeInfo);
    frame_t offset = 0;

    // Check if the action is due to start in the current block

    if (action->done_frames == 0)
    {
      // This action has not yet been "active"

      PaTime diff = action->requested_time - time;
      if (diff >= 0.0)
      {
        offset = seconds2samples(diff, state->samplerate);
        if (offset >= frameCount)
        {
          // We are too early, let's continue in the next block!

          // Due to inaccuracies in timeInfo, "diff" might have a small negative
          // value in a future block.  We don't count this as "belated" though:
          action->actual_time = -1.0;
          actionaddr = &(action->next);
          continue;
        }
        // Re-calculate "diff" to propagate rounding errors
        action->actual_time = time + (double)offset / state->samplerate;
      }
      else
      {
        // We are too late!
        if (action->actual_time == 0.0)
        {
          // allow_belated == False
          remove_action(actionaddr, state);
          continue;
        }
        action->actual_time = time;
      }
    }

    // Handle CANCEL action

    if (action->type == CANCEL)
    {
      // Since CANCEL actions are inserted in the beginning,
      // we need to search only the following list items
      for (struct action** i = &(action->next); *i; i = &((*i)->next))
      {
        if (*i == action->action)
        {
          struct action* delinquent = *i;

          if (delinquent->done_frames == 0)
          {
            // delinquent is not yet playing/recording

            frame_t delinquent_offset = 0;
            PaTime diff = delinquent->requested_time - time;
            if (diff >= 0.0)
            {
              delinquent_offset = seconds2samples(diff, state->samplerate);
              if (delinquent_offset >= offset)
              {
                // Removal is scheduled before playback/recording begins

                // TODO: save some more status information?
                delinquent->total_frames = 0;
                remove_action(i, state);
                break;
              }
            }
            else
            {
              if (delinquent->actual_time == 0.0)
              {
                // allow_belated == False

                // TODO: save some status information?
                break;  // The action will not be started, no need to cancel it
              }
            }

            if (delinquent->total_frames == ULONG_MAX
             || delinquent->total_frames + delinquent_offset > offset)
            {
              delinquent->total_frames = offset - delinquent_offset;
            }
            else
            {
              // TODO: stops on its own ... save some status information?
            }
          }
          else
          {
            CALLBACK_ASSERT(
                delinquent->total_frames >= delinquent->done_frames);
            if (delinquent->total_frames - delinquent->done_frames > offset)
            {
              delinquent->total_frames = delinquent->done_frames + offset;
            }
            else
            {
              // TODO: stops on its own ... save some status information?
            }
          }
          // TODO: save some informations to action->...?

          break;  // We found the action, no need to keep searching
        }
      }
      // TODO: what if the action to cancel wasn't found?

      remove_action(actionaddr, state);  // Remove the CANCEL action itself
      continue;
    }

    // Handle FETCH_AND_RESET_STATS action

    if (action->type == FETCH_AND_RESET_STATS)
    {
      action->stats = state->stats;
      state->stats = EMPTY_STATS;
      remove_action(actionaddr, state);
      continue;
    }

    // Store buffer over-/underflow information etc.

    get_stats(frameCount, statusFlags, &(action->stats));

    // Get number of remaining frames in the current block

    CALLBACK_ASSERT(action->total_frames >= action->done_frames);
    frame_t frames = action->total_frames - action->done_frames;

    if (frameCount < frames)
    {
      frames = frameCount;
    }
    if (frames + offset > frameCount)
    {
      CALLBACK_ASSERT(frameCount > offset);
      frames = frameCount - offset;
    }

    // Shove audio data around

    float* device_data = NULL;
    if (action->type == PLAY_BUFFER || action->type == PLAY_RINGBUFFER)
    {
      device_data = (float*)output + offset * state->output_channels;
    }
    else
    {
      CALLBACK_ASSERT(action->type == RECORD_BUFFER
                   || action->type == RECORD_RINGBUFFER);
      device_data = (float*) input + offset * state->input_channels;
    }

    if (action->type == PLAY_BUFFER || action->type == RECORD_BUFFER)
    {
      float* buffer = action->buffer + action->done_frames * action->channels;
      action->done_frames += frames;
      if (action->type == PLAY_BUFFER)
      {
        while (frames--)
        {
          for (frame_t c = 0; c < action->channels; c++)
          {
            CALLBACK_ASSERT(action->mapping[c] >= 1);
            CALLBACK_ASSERT(action->mapping[c] <= state->output_channels);
            device_data[action->mapping[c] - 1] += *buffer++;
          }
          device_data += state->output_channels;
        }
      }
      else
      {
        CALLBACK_ASSERT(action->type == RECORD_BUFFER);

        while (frames--)
        {
          for (frame_t c = 0; c < action->channels; c++)
          {
            CALLBACK_ASSERT(action->mapping[c] >= 1);
            CALLBACK_ASSERT(action->mapping[c] <= state->input_channels);
            *buffer++ = device_data[action->mapping[c] - 1];
          }
          device_data += state->input_channels;
        }
      }
    }
    else
    {
      CALLBACK_ASSERT(action->type ==   PLAY_RINGBUFFER
                   || action->type == RECORD_RINGBUFFER);

      float* block1 = NULL;
      float* block2 = NULL;
      ring_buffer_size_t size1 = 0;
      ring_buffer_size_t size2 = 0;
      ring_buffer_size_t totalsize = 0;

      if (action->type == PLAY_RINGBUFFER)
      {
        totalsize = PaUtil_GetRingBufferReadRegions(action->ringbuffer
          , (ring_buffer_size_t)frames
          , (void**)&block1, &size1, (void**)&block2, &size2);
        CALLBACK_ASSERT(!totalsize || size1);

        while (size1--)
        {
          for (frame_t c = 0; c < action->channels; c++)
          {
            CALLBACK_ASSERT(action->mapping[c] >= 1);
            CALLBACK_ASSERT(action->mapping[c] <= state->output_channels);
            device_data[action->mapping[c] - 1] += *block1++;
          }
          device_data += state->output_channels;
        }
        while (size2--)
        {
          for (frame_t c = 0; c < action->channels; c++)
          {
            device_data[action->mapping[c] - 1] += *block2++;
          }
          device_data += state->output_channels;
        }
        action->done_frames += (frame_t)totalsize;
        PaUtil_AdvanceRingBufferReadIndex(action->ringbuffer, totalsize);
      }
      else
      {
        CALLBACK_ASSERT(action->type == RECORD_RINGBUFFER);

        totalsize = PaUtil_GetRingBufferWriteRegions(action->ringbuffer
          , (ring_buffer_size_t)frames
          , (void**)&block1, &size1, (void**)&block2, &size2);
        CALLBACK_ASSERT(!totalsize || size1);

        while (size1--)
        {
          for (frame_t c = 0; c < action->channels; c++)
          {
            CALLBACK_ASSERT(action->mapping[c] >= 1);
            CALLBACK_ASSERT(action->mapping[c] <= state->input_channels);
            *block1++ = device_data[action->mapping[c] - 1];
          }
          device_data += state->input_channels;
        }
        while (size2--)
        {
          for (frame_t c = 0; c < action->channels; c++)
          {
            *block2++ = device_data[action->mapping[c] - 1];
          }
          device_data += state->input_channels;
        }
        action->done_frames += (frame_t)totalsize;
        PaUtil_AdvanceRingBufferWriteIndex(action->ringbuffer, totalsize);
      }

      if (totalsize < (ring_buffer_size_t)frames)
      {
        // Ring buffer is empty or full
        remove_action(actionaddr, state);
        continue;
      }
    }

    // Clean up, prepare next iteration

    if (action->done_frames == action->total_frames)
    {
      remove_action(actionaddr, state);
      continue;
    }
    actionaddr = &(action->next);
  }
  return paContinue;
}


/************************************************************/

static void *_cffi_types[] = {
/*  0 */ _CFFI_OP(_CFFI_OP_FUNCTION, 50), // int()(void const *, void *, unsigned long, PaStreamCallbackTimeInfo const *, unsigned long, void *)
/*  1 */ _CFFI_OP(_CFFI_OP_POINTER, 56), // void const *
/*  2 */ _CFFI_OP(_CFFI_OP_POINTER, 56), // void *
/*  3 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 10), // unsigned long
/*  4 */ _CFFI_OP(_CFFI_OP_POINTER, 42), // PaStreamCallbackTimeInfo const *
/*  5 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 10),
/*  6 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/*  7 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  8 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, long)
/*  9 */ _CFFI_OP(_CFFI_OP_POINTER, 43), // PaUtilRingBuffer *
/* 10 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9), // long
/* 11 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 12 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, long, long, void *)
/* 13 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 14 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 15 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 16 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 17 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 18 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, long, void * *, long *, void * *, long *)
/* 19 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 20 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 21 */ _CFFI_OP(_CFFI_OP_POINTER, 2), // void * *
/* 22 */ _CFFI_OP(_CFFI_OP_POINTER, 10), // long *
/* 23 */ _CFFI_OP(_CFFI_OP_NOOP, 21),
/* 24 */ _CFFI_OP(_CFFI_OP_NOOP, 22),
/* 25 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 26 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, void *, long)
/* 27 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 28 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 29 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 30 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 31 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer *, void const *, long)
/* 32 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 33 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/* 34 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 9),
/* 35 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 36 */ _CFFI_OP(_CFFI_OP_FUNCTION, 10), // long()(PaUtilRingBuffer const *)
/* 37 */ _CFFI_OP(_CFFI_OP_POINTER, 43), // PaUtilRingBuffer const *
/* 38 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 39 */ _CFFI_OP(_CFFI_OP_FUNCTION, 56), // void()(PaUtilRingBuffer *)
/* 40 */ _CFFI_OP(_CFFI_OP_NOOP, 9),
/* 41 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 42 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 0), // PaStreamCallbackTimeInfo
/* 43 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 1), // PaUtilRingBuffer
/* 44 */ _CFFI_OP(_CFFI_OP_POINTER, 45), // char *
/* 45 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 2), // char
/* 46 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14), // double
/* 47 */ _CFFI_OP(_CFFI_OP_ENUM, 0), // enum actiontype
/* 48 */ _CFFI_OP(_CFFI_OP_POINTER, 49), // float *
/* 49 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13), // float
/* 50 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 7), // int
/* 51 */ _CFFI_OP(_CFFI_OP_POINTER, 52), // struct action *
/* 52 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 2), // struct action
/* 53 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 3), // struct state
/* 54 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 4), // struct stats
/* 55 */ _CFFI_OP(_CFFI_OP_OPEN_ARRAY, 3), // unsigned long[]
/* 56 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 0), // void
};

static int _cffi_const_PLAY_BUFFER(unsigned long long *o)
{
  int n = (PLAY_BUFFER) <= 0;
  *o = (unsigned long long)((PLAY_BUFFER) | 0);  /* check that PLAY_BUFFER is an integer */
  return n;
}

static int _cffi_const_PLAY_RINGBUFFER(unsigned long long *o)
{
  int n = (PLAY_RINGBUFFER) <= 0;
  *o = (unsigned long long)((PLAY_RINGBUFFER) | 0);  /* check that PLAY_RINGBUFFER is an integer */
  return n;
}

static int _cffi_const_RECORD_BUFFER(unsigned long long *o)
{
  int n = (RECORD_BUFFER) <= 0;
  *o = (unsigned long long)((RECORD_BUFFER) | 0);  /* check that RECORD_BUFFER is an integer */
  return n;
}

static int _cffi_const_RECORD_RINGBUFFER(unsigned long long *o)
{
  int n = (RECORD_RINGBUFFER) <= 0;
  *o = (unsigned long long)((RECORD_RINGBUFFER) | 0);  /* check that RECORD_RINGBUFFER is an integer */
  return n;
}

static int _cffi_const_CANCEL(unsigned long long *o)
{
  int n = (CANCEL) <= 0;
  *o = (unsigned long long)((CANCEL) | 0);  /* check that CANCEL is an integer */
  return n;
}

static int _cffi_const_FETCH_AND_RESET_STATS(unsigned long long *o)
{
  int n = (FETCH_AND_RESET_STATS) <= 0;
  *o = (unsigned long long)((FETCH_AND_RESET_STATS) | 0);  /* check that FETCH_AND_RESET_STATS is an integer */
  return n;
}

static long _cffi_d_PaUtil_AdvanceRingBufferReadIndex(PaUtilRingBuffer * x0, long x1)
{
  return PaUtil_AdvanceRingBufferReadIndex(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_AdvanceRingBufferReadIndex(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "PaUtil_AdvanceRingBufferReadIndex", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_AdvanceRingBufferReadIndex(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_AdvanceRingBufferReadIndex _cffi_d_PaUtil_AdvanceRingBufferReadIndex
#endif

static long _cffi_d_PaUtil_AdvanceRingBufferWriteIndex(PaUtilRingBuffer * x0, long x1)
{
  return PaUtil_AdvanceRingBufferWriteIndex(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_AdvanceRingBufferWriteIndex(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "PaUtil_AdvanceRingBufferWriteIndex", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_AdvanceRingBufferWriteIndex(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_AdvanceRingBufferWriteIndex _cffi_d_PaUtil_AdvanceRingBufferWriteIndex
#endif

static void _cffi_d_PaUtil_FlushRingBuffer(PaUtilRingBuffer * x0)
{
  PaUtil_FlushRingBuffer(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_FlushRingBuffer(PyObject *self, PyObject *arg0)
{
  PaUtilRingBuffer * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { PaUtil_FlushRingBuffer(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_PaUtil_FlushRingBuffer _cffi_d_PaUtil_FlushRingBuffer
#endif

static long _cffi_d_PaUtil_GetRingBufferReadAvailable(PaUtilRingBuffer const * x0)
{
  return PaUtil_GetRingBufferReadAvailable(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferReadAvailable(PyObject *self, PyObject *arg0)
{
  PaUtilRingBuffer const * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(37), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(37), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferReadAvailable(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_GetRingBufferReadAvailable _cffi_d_PaUtil_GetRingBufferReadAvailable
#endif

static long _cffi_d_PaUtil_GetRingBufferReadRegions(PaUtilRingBuffer * x0, long x1, void * * x2, long * x3, void * * x4, long * x5)
{
  return PaUtil_GetRingBufferReadRegions(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferReadRegions(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  void * * x2;
  long * x3;
  void * * x4;
  long * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "PaUtil_GetRingBufferReadRegions", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(21), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (long *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(22), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(21), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (long *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(22), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferReadRegions(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_GetRingBufferReadRegions _cffi_d_PaUtil_GetRingBufferReadRegions
#endif

static long _cffi_d_PaUtil_GetRingBufferWriteAvailable(PaUtilRingBuffer const * x0)
{
  return PaUtil_GetRingBufferWriteAvailable(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferWriteAvailable(PyObject *self, PyObject *arg0)
{
  PaUtilRingBuffer const * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(37), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(37), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferWriteAvailable(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_GetRingBufferWriteAvailable _cffi_d_PaUtil_GetRingBufferWriteAvailable
#endif

static long _cffi_d_PaUtil_GetRingBufferWriteRegions(PaUtilRingBuffer * x0, long x1, void * * x2, long * x3, void * * x4, long * x5)
{
  return PaUtil_GetRingBufferWriteRegions(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_GetRingBufferWriteRegions(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  void * * x2;
  long * x3;
  void * * x4;
  long * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "PaUtil_GetRingBufferWriteRegions", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(21), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (long *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(22), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(21), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(21), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(22), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (long *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(22), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_GetRingBufferWriteRegions(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_GetRingBufferWriteRegions _cffi_d_PaUtil_GetRingBufferWriteRegions
#endif

static long _cffi_d_PaUtil_InitializeRingBuffer(PaUtilRingBuffer * x0, long x1, long x2, void * x3)
{
  return PaUtil_InitializeRingBuffer(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_InitializeRingBuffer(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  long x1;
  long x2;
  void * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "PaUtil_InitializeRingBuffer", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, long);
  if (x1 == (long)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, long);
  if (x2 == (long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_InitializeRingBuffer(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_InitializeRingBuffer _cffi_d_PaUtil_InitializeRingBuffer
#endif

static long _cffi_d_PaUtil_ReadRingBuffer(PaUtilRingBuffer * x0, void * x1, long x2)
{
  return PaUtil_ReadRingBuffer(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_ReadRingBuffer(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  void * x1;
  long x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "PaUtil_ReadRingBuffer", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, long);
  if (x2 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_ReadRingBuffer(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_ReadRingBuffer _cffi_d_PaUtil_ReadRingBuffer
#endif

static long _cffi_d_PaUtil_WriteRingBuffer(PaUtilRingBuffer * x0, void const * x1, long x2)
{
  return PaUtil_WriteRingBuffer(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_PaUtil_WriteRingBuffer(PyObject *self, PyObject *args)
{
  PaUtilRingBuffer * x0;
  void const * x1;
  long x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  long result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "PaUtil_WriteRingBuffer", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(9), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (PaUtilRingBuffer *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(9), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, long);
  if (x2 == (long)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = PaUtil_WriteRingBuffer(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, long);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_PaUtil_WriteRingBuffer _cffi_d_PaUtil_WriteRingBuffer
#endif

static int _cffi_d_callback(void const * x0, void * x1, unsigned long x2, PaStreamCallbackTimeInfo const * x3, unsigned long x4, void * x5)
{
  return callback(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_callback(PyObject *self, PyObject *args)
{
  void const * x0;
  void * x1;
  unsigned long x2;
  PaStreamCallbackTimeInfo const * x3;
  unsigned long x4;
  void * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  int result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "callback", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, unsigned long);
  if (x2 == (unsigned long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(4), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (PaStreamCallbackTimeInfo const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(4), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, unsigned long);
  if (x4 == (unsigned long)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = callback(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, int);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_callback _cffi_d_callback
#endif

static int _cffi_const_ULONG_MAX(unsigned long long *o)
{
  int n = (ULONG_MAX) <= 0;
  *o = (unsigned long long)((ULONG_MAX) | 0);  /* check that ULONG_MAX is an integer */
  return n;
}

_CFFI_UNUSED_FN
static void _cffi_checkfld__PaUtilRingBuffer(PaUtilRingBuffer *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->bufferSize) | 0);  /* check that 'PaUtilRingBuffer.bufferSize' is an integer */
  (void)((p->writeIndex) | 0);  /* check that 'PaUtilRingBuffer.writeIndex' is an integer */
  (void)((p->readIndex) | 0);  /* check that 'PaUtilRingBuffer.readIndex' is an integer */
  (void)((p->bigMask) | 0);  /* check that 'PaUtilRingBuffer.bigMask' is an integer */
  (void)((p->smallMask) | 0);  /* check that 'PaUtilRingBuffer.smallMask' is an integer */
  (void)((p->elementSizeBytes) | 0);  /* check that 'PaUtilRingBuffer.elementSizeBytes' is an integer */
  { char * *tmp = &p->buffer; (void)tmp; }
}
struct _cffi_align__PaUtilRingBuffer { char x; PaUtilRingBuffer y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld_struct_action(struct action *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  { enum actiontype const *tmp = &p->type; (void)tmp; }
  { double const *tmp = &p->requested_time; (void)tmp; }
  { double *tmp = &p->actual_time; (void)tmp; }
  { struct action * *tmp = &p->next; (void)tmp; }
  { float * const *tmp = &p->buffer; (void)tmp; }
  { PaUtilRingBuffer * const *tmp = &p->ringbuffer; (void)tmp; }
  { struct action * const *tmp = &p->action; (void)tmp; }
  (void)((p->total_frames) | 0);  /* check that 'struct action.total_frames' is an integer */
  (void)((p->done_frames) | 0);  /* check that 'struct action.done_frames' is an integer */
  { struct stats *tmp = &p->stats; (void)tmp; }
  (void)((p->channels) | 0);  /* check that 'struct action.channels' is an integer */
  { unsigned long const *tmp = &p->mapping[0]; (void)tmp; }
}
struct _cffi_align_struct_action { char x; struct action y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld_struct_state(struct state *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->input_channels) | 0);  /* check that 'struct state.input_channels' is an integer */
  (void)((p->output_channels) | 0);  /* check that 'struct state.output_channels' is an integer */
  { double *tmp = &p->samplerate; (void)tmp; }
  { PaUtilRingBuffer * const *tmp = &p->action_q; (void)tmp; }
  { PaUtilRingBuffer * const *tmp = &p->result_q; (void)tmp; }
  { struct action * *tmp = &p->actions; (void)tmp; }
  { struct stats *tmp = &p->stats; (void)tmp; }
}
struct _cffi_align_struct_state { char x; struct state y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld_struct_stats(struct stats *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  (void)((p->blocks) | 0);  /* check that 'struct stats.blocks' is an integer */
  (void)((p->min_blocksize) | 0);  /* check that 'struct stats.min_blocksize' is an integer */
  (void)((p->max_blocksize) | 0);  /* check that 'struct stats.max_blocksize' is an integer */
  (void)((p->input_underflows) | 0);  /* check that 'struct stats.input_underflows' is an integer */
  (void)((p->input_overflows) | 0);  /* check that 'struct stats.input_overflows' is an integer */
  (void)((p->output_underflows) | 0);  /* check that 'struct stats.output_underflows' is an integer */
  (void)((p->output_overflows) | 0);  /* check that 'struct stats.output_overflows' is an integer */
}
struct _cffi_align_struct_stats { char x; struct stats y; };

static const struct _cffi_global_s _cffi_globals[] = {
  { "CANCEL", (void *)_cffi_const_CANCEL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "FETCH_AND_RESET_STATS", (void *)_cffi_const_FETCH_AND_RESET_STATS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "PLAY_BUFFER", (void *)_cffi_const_PLAY_BUFFER, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "PLAY_RINGBUFFER", (void *)_cffi_const_PLAY_RINGBUFFER, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "PaUtil_AdvanceRingBufferReadIndex", (void *)_cffi_f_PaUtil_AdvanceRingBufferReadIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 8), (void *)_cffi_d_PaUtil_AdvanceRingBufferReadIndex },
  { "PaUtil_AdvanceRingBufferWriteIndex", (void *)_cffi_f_PaUtil_AdvanceRingBufferWriteIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 8), (void *)_cffi_d_PaUtil_AdvanceRingBufferWriteIndex },
  { "PaUtil_FlushRingBuffer", (void *)_cffi_f_PaUtil_FlushRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 39), (void *)_cffi_d_PaUtil_FlushRingBuffer },
  { "PaUtil_GetRingBufferReadAvailable", (void *)_cffi_f_PaUtil_GetRingBufferReadAvailable, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 36), (void *)_cffi_d_PaUtil_GetRingBufferReadAvailable },
  { "PaUtil_GetRingBufferReadRegions", (void *)_cffi_f_PaUtil_GetRingBufferReadRegions, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 18), (void *)_cffi_d_PaUtil_GetRingBufferReadRegions },
  { "PaUtil_GetRingBufferWriteAvailable", (void *)_cffi_f_PaUtil_GetRingBufferWriteAvailable, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 36), (void *)_cffi_d_PaUtil_GetRingBufferWriteAvailable },
  { "PaUtil_GetRingBufferWriteRegions", (void *)_cffi_f_PaUtil_GetRingBufferWriteRegions, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 18), (void *)_cffi_d_PaUtil_GetRingBufferWriteRegions },
  { "PaUtil_InitializeRingBuffer", (void *)_cffi_f_PaUtil_InitializeRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 12), (void *)_cffi_d_PaUtil_InitializeRingBuffer },
  { "PaUtil_ReadRingBuffer", (void *)_cffi_f_PaUtil_ReadRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 26), (void *)_cffi_d_PaUtil_ReadRingBuffer },
  { "PaUtil_WriteRingBuffer", (void *)_cffi_f_PaUtil_WriteRingBuffer, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 31), (void *)_cffi_d_PaUtil_WriteRingBuffer },
  { "RECORD_BUFFER", (void *)_cffi_const_RECORD_BUFFER, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "RECORD_RINGBUFFER", (void *)_cffi_const_RECORD_RINGBUFFER, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "ULONG_MAX", (void *)_cffi_const_ULONG_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "callback", (void *)_cffi_f_callback, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_callback },
};

static const struct _cffi_field_s _cffi_fields[] = {
  { "bufferSize", offsetof(PaUtilRingBuffer, bufferSize),
                  sizeof(((PaUtilRingBuffer *)0)->bufferSize),
                  _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "writeIndex", offsetof(PaUtilRingBuffer, writeIndex),
                  sizeof(((PaUtilRingBuffer *)0)->writeIndex),
                  _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "readIndex", offsetof(PaUtilRingBuffer, readIndex),
                 sizeof(((PaUtilRingBuffer *)0)->readIndex),
                 _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "bigMask", offsetof(PaUtilRingBuffer, bigMask),
               sizeof(((PaUtilRingBuffer *)0)->bigMask),
               _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "smallMask", offsetof(PaUtilRingBuffer, smallMask),
                 sizeof(((PaUtilRingBuffer *)0)->smallMask),
                 _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "elementSizeBytes", offsetof(PaUtilRingBuffer, elementSizeBytes),
                        sizeof(((PaUtilRingBuffer *)0)->elementSizeBytes),
                        _CFFI_OP(_CFFI_OP_NOOP, 10) },
  { "buffer", offsetof(PaUtilRingBuffer, buffer),
              sizeof(((PaUtilRingBuffer *)0)->buffer),
              _CFFI_OP(_CFFI_OP_NOOP, 44) },
  { "type", offsetof(struct action, type),
            sizeof(((struct action *)0)->type),
            _CFFI_OP(_CFFI_OP_NOOP, 47) },
  { "requested_time", offsetof(struct action, requested_time),
                      sizeof(((struct action *)0)->requested_time),
                      _CFFI_OP(_CFFI_OP_NOOP, 46) },
  { "actual_time", offsetof(struct action, actual_time),
                   sizeof(((struct action *)0)->actual_time),
                   _CFFI_OP(_CFFI_OP_NOOP, 46) },
  { "next", offsetof(struct action, next),
            sizeof(((struct action *)0)->next),
            _CFFI_OP(_CFFI_OP_NOOP, 51) },
  { "buffer", offsetof(struct action, buffer),
              sizeof(((struct action *)0)->buffer),
              _CFFI_OP(_CFFI_OP_NOOP, 48) },
  { "ringbuffer", offsetof(struct action, ringbuffer),
                  sizeof(((struct action *)0)->ringbuffer),
                  _CFFI_OP(_CFFI_OP_NOOP, 9) },
  { "action", offsetof(struct action, action),
              sizeof(((struct action *)0)->action),
              _CFFI_OP(_CFFI_OP_NOOP, 51) },
  { "total_frames", offsetof(struct action, total_frames),
                    sizeof(((struct action *)0)->total_frames),
                    _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "done_frames", offsetof(struct action, done_frames),
                   sizeof(((struct action *)0)->done_frames),
                   _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "stats", offsetof(struct action, stats),
             sizeof(((struct action *)0)->stats),
             _CFFI_OP(_CFFI_OP_NOOP, 54) },
  { "channels", offsetof(struct action, channels),
                sizeof(((struct action *)0)->channels),
                _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "mapping", offsetof(struct action, mapping),
               (size_t)-1,
               _CFFI_OP(_CFFI_OP_NOOP, 55) },
  { "input_channels", offsetof(struct state, input_channels),
                      sizeof(((struct state *)0)->input_channels),
                      _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "output_channels", offsetof(struct state, output_channels),
                       sizeof(((struct state *)0)->output_channels),
                       _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "samplerate", offsetof(struct state, samplerate),
                  sizeof(((struct state *)0)->samplerate),
                  _CFFI_OP(_CFFI_OP_NOOP, 46) },
  { "action_q", offsetof(struct state, action_q),
                sizeof(((struct state *)0)->action_q),
                _CFFI_OP(_CFFI_OP_NOOP, 9) },
  { "result_q", offsetof(struct state, result_q),
                sizeof(((struct state *)0)->result_q),
                _CFFI_OP(_CFFI_OP_NOOP, 9) },
  { "actions", offsetof(struct state, actions),
               sizeof(((struct state *)0)->actions),
               _CFFI_OP(_CFFI_OP_NOOP, 51) },
  { "stats", offsetof(struct state, stats),
             sizeof(((struct state *)0)->stats),
             _CFFI_OP(_CFFI_OP_NOOP, 54) },
  { "blocks", offsetof(struct stats, blocks),
              sizeof(((struct stats *)0)->blocks),
              _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "min_blocksize", offsetof(struct stats, min_blocksize),
                     sizeof(((struct stats *)0)->min_blocksize),
                     _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "max_blocksize", offsetof(struct stats, max_blocksize),
                     sizeof(((struct stats *)0)->max_blocksize),
                     _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "input_underflows", offsetof(struct stats, input_underflows),
                        sizeof(((struct stats *)0)->input_underflows),
                        _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "input_overflows", offsetof(struct stats, input_overflows),
                       sizeof(((struct stats *)0)->input_overflows),
                       _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "output_underflows", offsetof(struct stats, output_underflows),
                         sizeof(((struct stats *)0)->output_underflows),
                         _CFFI_OP(_CFFI_OP_NOOP, 3) },
  { "output_overflows", offsetof(struct stats, output_overflows),
                        sizeof(((struct stats *)0)->output_overflows),
                        _CFFI_OP(_CFFI_OP_NOOP, 3) },
};

static const struct _cffi_struct_union_s _cffi_struct_unions[] = {
  { "PaStreamCallbackTimeInfo", 42, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "PaUtilRingBuffer", 43, _CFFI_F_CHECK_FIELDS,
    sizeof(PaUtilRingBuffer), offsetof(struct _cffi_align__PaUtilRingBuffer, y), 0, 7 },
  { "action", 52, 0,
    sizeof(struct action), offsetof(struct _cffi_align_struct_action, y), 7, 12 },
  { "state", 53, _CFFI_F_CHECK_FIELDS,
    sizeof(struct state), offsetof(struct _cffi_align_struct_state, y), 19, 7 },
  { "stats", 54, _CFFI_F_CHECK_FIELDS,
    sizeof(struct stats), offsetof(struct _cffi_align_struct_stats, y), 26, 7 },
};

static const struct _cffi_enum_s _cffi_enums[] = {
  { "actiontype", 47, _cffi_prim_int(sizeof(enum actiontype), ((enum actiontype)-1) <= 0),
    "PLAY_BUFFER,PLAY_RINGBUFFER,RECORD_BUFFER,RECORD_RINGBUFFER,CANCEL,FETCH_AND_RESET_STATS" },
};

static const struct _cffi_typename_s _cffi_typenames[] = {
  { "PaStreamCallbackFlags", 3 },
  { "PaStreamCallbackTimeInfo", 42 },
  { "PaTime", 46 },
  { "PaUtilRingBuffer", 43 },
  { "frame_t", 3 },
  { "ring_buffer_size_t", 10 },
};

static const struct _cffi_type_context_s _cffi_type_context = {
  _cffi_types,
  _cffi_globals,
  _cffi_fields,
  _cffi_struct_unions,
  _cffi_enums,
  _cffi_typenames,
  18,  /* num_globals */
  5,  /* num_struct_unions */
  1,  /* num_enums */
  6,  /* num_typenames */
  NULL,  /* no includes */
  57,  /* num_types */
  0,  /* flags */
};

#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit__rtmixer(const void *p[])
{
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
#if PY_MAJOR_VERSION >= 3
    return NULL;
#endif
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit__rtmixer(void) { return NULL; }
#  else
     init_rtmixer(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit__rtmixer(void)
{
  return _cffi_init("_rtmixer", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
init_rtmixer(void)
{
  _cffi_init("_rtmixer", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif
